package oqbe.actions.development;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;

import oqbe.exporters.ASTExporter;
import oqbe.exporters.OQBEModel;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.IDE;

import vide.ast.abstracts.AbstractNode;
import vide.converters.VideCodeToAST;

public class TestRunVEB implements IWorkbenchWindowActionDelegate {

	protected class Alias {
		String name;
		String value;
		public Alias(String name, String value) {
			this.name = name;
			this.value = value;
		}
	}

	protected IWorkbenchWindow window;
	protected String workspaceDir;
	protected BufferedReader reader;
	protected StringBuffer logBuffer;
	protected int testCount;
	protected int successCount;
	protected String testsProject = "VEB Tests";	
	protected EList<Alias> aliases;
	
	public void dispose() {
	}

	public void init(IWorkbenchWindow window) {
		this.window = window;
		workspaceDir = ResourcesPlugin.getWorkspace().getRoot().getLocation().toString();		
	}

	public void run(IAction action) {
		testCount = 0;
		successCount = 0;
		logBuffer = new StringBuffer();
		aliases = new BasicEList<Alias>();
		
		String inputFileName = workspaceDir + "/" + testsProject + "/testsDef.txt";
		try {
			try {
				openFile(inputFileName);
			} catch (Exception e) {
				MessageDialog.openError(window.getShell(), 
						"Error executing VEB tests!", 
						"Tests definition file (/VEB Tests/TestsDef.txt) not found!");
				throw e;
			}
			try {
				boolean endOfAliasList = false;
				String line;
				String[] parts = null;
				while ( (!endOfAliasList) && ((line = readFileLine()) != null) && (line.trim().length()> 0) ) {
					parts = line.split("=");
					if (parts.length < 2)
						endOfAliasList = true;
					else
						aliases.add(new Alias(parts[0], parts[1]));
				}
				
				String fileName = null;
				String code = "";						
				while ((line = readFileLine()) != null) {
					if (line.trim().length() != 0) {
						if (fileName == null)
							fileName = line.trim();
						else {
							if (code.length() > 0)
								code += "\n";
							code += line;
						}
					} else if (code.length() > 0) {
						executeTest(fileName, code);
						fileName = null;
						code = "";
					}
				}
				if (code.length() != 0)
					executeTest(fileName, code);
			} finally {
				closeFile();
			}
			
			String outputFileName = workspaceDir + "/" + testsProject + "/results.txt";
			String results = String.format("Test count: %d\nPassed tests: %d\nErrors: %d\nRatio: %.0f%s",
					new Object[] {
					testCount, successCount, testCount - successCount,
					(float) successCount / testCount * 100, "%"
					}); 
			writeFile(outputFileName, results + "\n" + logBuffer.toString());
			openFileInEditor(outputFileName);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void selectionChanged(IAction action, ISelection selection) {
	}

	//
	
	protected void executeTest(String fileName, String code) {
		boolean srcFileNotFound = false;
		fileName = replaceAliases(fileName);
		testCount ++;
		String s = "\n" + fileName + "... ";
		String sourceCode = "";
		try {
			// reformat source code
			try {
				AbstractNode ast = VideCodeToAST.convert(code);
				sourceCode = ast.asCode();
				if (sourceCode.endsWith("\n"))
					sourceCode = sourceCode.substring(0, sourceCode.length() - 1);
			} catch (Exception e) {
				sourceCode = code;
				throw e;
			}
			
			// generate expression code
			String dirSep = fileName.startsWith("/") ? "" : "/";
			OQBEModel model = null;
			try {
				model  = new OQBEModel(new Path(workspaceDir + dirSep + fileName));
			} catch (Exception e) {
				if (e.getCause() instanceof FileNotFoundException)
					srcFileNotFound = true;				
				throw e;
			}
			ASTExporter generator = new ASTExporter();
			AbstractNode node = generator.generate(model);
			String generatedCode = node.asCode();			
					
			// compare source and generated code
			if (("" + sourceCode).equals(generatedCode)) {
				s += "OK";
				successCount ++;
			} else {
				s += "test failed.";
				s += "\nSource code:\n" + sourceCode;
				s += "\nGenerated code:\n" + generatedCode;
			}
		} catch (Exception e) {			
			s += "test failed";
			if (srcFileNotFound)
				s += " - source file not found!";
			else {
				s += ".";
				s += "\nSource code:\n" + sourceCode;			
				s += "\nException occured:" + getStackTrace(e);
			}
		}
		logBuffer.append(s);
	}
	
	// reading file
	
    protected void openFile(String fileName) throws FileNotFoundException {
    	File file = new File(fileName);
		reader = new BufferedReader (new FileReader(file));
    }
    
    protected String readFileLine() throws IOException {
    	return reader.readLine();
    }
    
    protected void closeFile() throws IOException {
    	reader.close();
    }
    
    // writing file
    
    protected void writeFile(String fileName, String dataString) {
    	File file = new File(fileName);
    	try {
    		PrintWriter out = new PrintWriter (new BufferedWriter (new FileWriter (file)));
    		out.print(dataString);
    		out.flush();
    		out.close();
    	} catch (IOException e) {
    		e.printStackTrace();
    	} 
    }	
        
    // opening file in default editor
    
    protected void openFileInEditor(String fileName) {
		final String fileToOpen = fileName;
		
		window.getShell().getDisplay().asyncExec(new Runnable() {
			public void run() {
				IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
				try {
					IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();					
					final IFile file = root.findFilesForLocation(new Path(fileToOpen))[0];
					file.refreshLocal(IResource.DEPTH_INFINITE, null);
					IDE.openEditor(page, file, true);
				} catch (Exception e) {
				}
			}
		});
	}

    //
    
    protected String getStackTrace(Exception e) {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        pw.close();
        return sw.toString();
    }    	

    //
    
    protected String replaceAliases(String s) {
    	for (Alias a : aliases)
    		s = s.replace(a.name, a.value);
    	return s;
    }
    
}