package oqbe.exporters;

import java.util.ArrayList;
import java.util.HashMap;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.uml2.uml.Element;

import vide.ast.abstracts.AbstractNode;
import vide.ast.declarations.VariableDeclarationCS;
import vide.ast.declarations.VariableDeclarationListCS;
import vide.ast.editor.statements.VideEditorCode;
import vide.ast.expressions.ArgumentsCS;
import vide.ast.expressions.AttributeCallCS;
import vide.ast.expressions.AttributeOrVariableCS;
import vide.ast.expressions.BinaryExpressionCS;
import vide.ast.expressions.CollectionOperationCS;
import vide.ast.expressions.OclExpressionCS;
import vide.ast.expressions.OperationCallCS;
import vide.ast.expressions.Operator;
import vide.ast.expressions.UnaryExpressionCS;
import vide.ast.expressions.literals.CollectionLiteralExpCS;
import vide.ast.expressions.literals.TupleLiteralExpCS;
import vide.ast.names.SimpleNameCS;
import vide.converters.VideCodeToAST;
import OQBE.Comparator;
import OQBE.Example;
import OQBE.LinkExample;
import OQBE.NestedOutput;
import OQBE.OclExpression;
import OQBE.Predicate;
import OQBE.SortFlag;
import OQBE.SortOrder;

public class ASTExporter {
	
	protected class GenOptions {
		public CallType callType = CallType.STANDARD;
		public LinkExample nestedLink = null;
		public String prevVarName = null;
		public GenOptions() {						
		}
		public GenOptions(LinkExample nestedLink, String prevVarName, CallType callType) {
			this.nestedLink = nestedLink;
			this.prevVarName = prevVarName;
			this.callType = callType;
		}
	}
	
	private OQBEModel model;
	private boolean debugMode = true;
	private OQBEIterator iterator;
	private TupleLiteralExpCS prevTuple;
	private AbstractNode startingExp;
	private Example firstExample = null;
	private OQBEItems<OQBEOutput> outputs;
	private HashMap<Example, String> fields;
	private OQBEItems<OQBESort> sorts;
	
	private GenOptions options = null;

	private static SimpleLogger logger = SimpleLogger.getInstance();

	//
	
	protected void initialize(OQBEModel model, GenOptions options) {
		this.model = model;
		this.options = options;
		
		outputs = new OQBEItems<OQBEOutput>(/*"outputs"*/);
		prevTuple = null;		
		fields = new HashMap<Example, String>();
		sorts = new OQBEItems<OQBESort>(/*"sorts"*/);
	}

	// generate

	public AbstractNode generate(OQBEModel model) throws Exception { 		
		return generate(model, /*null,*/ new GenOptions());
	}

	public AbstractNode generate(OQBEModel model, /*Example firstExample,*/ GenOptions options) throws Exception {
		if (options.callType != CallType.STANDARD)
			logger.incIndent();

		initialize(model, options);
		if (options.callType == CallType.STANDARD)
			firstExample = getFirstExample();
		else
			firstExample = options.nestedLink.getTarget();
		OclExpressionCS  result = generate();

		if (options.callType != CallType.STANDARD)
			logger.decIndent();
		
		return result;
	}
	
	private OclExpressionCS generate() throws Exception {
		// first expression
		OclExpressionCS result = generateFirstAttrOrVar(firstExample, options.nestedLink);
		startingExp = result;
		if ( (options.callType != CallType.FOR_ALL) && (firstExample.getPredicates().size() > 0) ) {
			processPredicates(firstExample, result); // outputs and sorts
			result = generateFilterExp(result, firstExample); // filter expression
		}		
		if (firstExample.getExpressions().size() > 0) {
			if (firstExample.getExpressions().size() > 0) {
				processOclExpressions(firstExample); // outputs and sorts
				result = generateOclExpression(result, firstExample); // filter expression
			}
		}
		
		// next expressions
		iterator = new OQBEIterator(model, firstExample, options.callType);
		Example example = iterator.getNext();
		while (example != null) {
			boolean alreadyVisited = iterator.alreadyVisited();
			LinkExample link = iterator.getCurrentLink();
			if (model.linkToNestedOutput(link))
				rememberNestedOutput(link);
			else
				result = generateCollWithLink(result, example, link, alreadyVisited);
			if (alreadyVisited)
				result = generateVisited(result, link);
			example = iterator.getNext();
		}
		
		// comparators
		result = generateComparators(result);
		// sort criteria
		result = generateSortExp(result);
		// nested outputs
		generateNestedOutputs();
		// final collect
		result = generateFinalCollect(result);
		
		return result;
	}
	
	private OclExpressionCS generateFirstAttrOrVar(Example example, LinkExample nestedLink) throws Exception {		
		logger.startOfMethod("generateFirstAttrOrVar");
			
		if ( (options.callType == CallType.NESTED_OUTPUT) && (options.nestedLink == null) )
			throw new Exception("Nested link not assigned in nested call!");
		if ( (options.callType == CallType.FOR_ALL) && (nestedLink == null) )
			throw new Exception("Link not assigned in nested for-all call!");

		OclExpressionCS result;
		if (options.callType == CallType.NESTED_OUTPUT) {
			// t.[v2].<link>
			Element tmpUmlElement = nestedLink.getDescribedBy();
			
			NotNullEList<String> attrs = new NotNullEList<String>();
			attrs.add("t");
			attrs.add(options.prevVarName);			
			attrs.add(getLinkName(nestedLink));
			
			result = ASTHelper.createVoAoAC(attrs, tmpUmlElement);
			result.umlElement = tmpUmlElement;

		} else if (options.callType == CallType.FOR_ALL) {
			// v1.product
			Element tmpUmlElement = nestedLink.getDescribedBy();
			AttributeOrVariableCS attr = ASTHelper.createAttrOrVar(getLinkName(nestedLink), tmpUmlElement);
			result = new AttributeCallCS(ASTHelper.createAttrOrVar(options.prevVarName, tmpUmlElement), attr);
		} else {
			// product
			// Product.allInstances()
			result = ASTHelper.createAttrOrVar(getExampleName(example), example.getDescribes());
			if ( (example.getExampleName() == null) || (example.getExampleName().length() == 0) )
				result = new OperationCallCS(result, ASTHelper.createSN("allInstances"), null);
		}
		if (example.getOutput() != null)
			addOutput(example, result, null, example.getOutput().getTupleFieldName(), false, null);
		
		logger.endOfMethod(result);
		return result;
	}		
	
	private OclExpressionCS generateFilterExp(OclExpressionCS baseExp, Example example, String expPrefix) throws Exception {
		logger.startOfMethod("generateFilterExp");
		
		OclExpressionCS result = baseExp;
		if (getNotFlaggedPredicates(example).size() > 0)
			// expressions prefix only for for-alls!
			result = generateSelectOp(baseExp, example, expPrefix, expPrefix == null ? false : true);
		
		logger.endOfMethod(result);
		return result;
	}
	
	private OclExpressionCS generateFilterExp(OclExpressionCS baseExp, Example example) throws Exception {
		return generateFilterExp(baseExp, example, null);
	}
	
	private OclExpressionCS generateSelectOp(OclExpressionCS baseExp, Example example, String expPrefix, boolean forAllExp) throws Exception {
		logger.startOfMethod("generateSelectOp");
		
		OclExpressionCS result = baseExp;
		
		// expression
		OclExpressionCS expression = null;
		EList<Predicate> predicates = getNotFlaggedPredicates(example);
		if (predicates.size() > 0)
			expression = ASTHelper.createBinExpTree(predicates, 0, expPrefix);
		
		// create select operation
		if (expression !=  null) {
			// method name		
			SimpleNameCS methodName = forAllExp ? ASTHelper.createSN("forAll") : ASTHelper.createSN("select");
			//
			result = new CollectionOperationCS(ASTHelper.packExpression(result), 
				methodName, expression, null);
		}
		
		logger.endOfMethod(result);
		return result;
	}
	
	private AttributeCallCS generateLinkExpression(LinkExample link) throws Exception {
		logger.startOfMethod("generateLinkExpression");
		
		AttributeCallCS result = null;
		String linkName = getLinkName(link);
		if (prevTuple == null)
			result = ASTHelper.createAttrCall(ASTHelper.createSN("v"), linkName, link.getDescribedBy());
		else {
			String prevVarName = fields.get(link.getSource());			
			result = ASTHelper.createAttrCall(ASTHelper.createSN("v"), prevVarName, linkName, link.getDescribedBy());
		}
		
		logger.endOfMethod(result);
		return result;		
	}
	
	private OclExpressionCS parseExp(OclExpression diagramExp) throws Exception {
		OclExpressionCS result = null;
		// parse expression text
		try {				
			String expText = diagramExp.getExpression();
			if (!expText.endsWith(";"))
				expText += ";";
			VideEditorCode editorCode = (VideEditorCode) VideCodeToAST.convert(expText);
			result = (OclExpressionCS) editorCode.statements.get(0);
		} catch (Exception e) {
			throw new Exception("Parsing failed!\n" + "OclExpression:\n" + diagramExp.getExpression());
		}
		return result;
	}
	
	private OclExpressionCS generateOclExpression(OclExpressionCS baseExp, Example example) throws Exception {
		logger.startOfMethod("generateOclExpression");
		
		OclExpressionCS result = baseExp;
		EList<OclExpression> expressions = getNotFlaggedOclExpressions(example);
		for (OclExpression diagramExp : expressions) {
			OclExpressionCS generatedExp = parseExp(diagramExp);	
			// create select operation
			SimpleNameCS methodName = ASTHelper.createSN("select");		
			result = new CollectionOperationCS(ASTHelper.packExpression(result),
					methodName, generatedExp, null);							
		}
		
		logger.endOfMethod(result);
		return result;
	}

	private OclExpressionCS generateCollWithLink(OclExpressionCS baseExp, 
			Example target, LinkExample link, boolean alreadyVisited) throws Exception {
		logger.startOfMethod("generateCollWithLink");
		
		if ( !((link == null) || (link.getTarget() == target)) ) 
			throw new Exception("Invalid parameters of " + getClass().getName() + ".generateCollectOp() method!");
		
		boolean linkToNestedOutput = model.linkToNestedOutput(link);
		boolean linkToForAll = model.linkToForAll(link);
			
		// ->collect(v | v.items->...) (collect with link)
		
		// 1. exp (base expression)
		
		OclExpressionCS exp = null;
		SimpleNameCS methodName = null;		
		ArgumentsCS args = null;
		
		if (linkToForAll)
			exp = baseExp;
		else {
			// iterator
			VariableDeclarationCS iteratorDecl = new VariableDeclarationCS(ASTHelper.createSN("v"));
			args = new ArgumentsCS(iteratorDecl);
			// collection method name 
			methodName = ASTHelper.createSN("collect");
			// expression			
			if (link != null)
				exp = generateLinkExpression(link);
			else
				exp = ASTHelper.createAttrOrVar(getExampleName(target), null);			
			logger.printLogLine("generated exp:\n" + exp.asExp());
			
			// 2. ->select (filter expression)
	
			if (!linkToNestedOutput)
				exp = generateFilterExp(exp, target);

			// 3. OCL expressions
			
			if (target.getExpressions().size() > 0) {
				processOclExpressions(target);
				exp = generateOclExpression(exp, target);
			}
		}
		
		// 4. collect with Tuple
		
		exp = generateCollWithTuple(exp, target, link, alreadyVisited);
	
		// 5. resulting collect 
				
		OclExpressionCS result;
		if (linkToForAll)
			result = generateForAll(exp, link);
		else
			result = new CollectionOperationCS(ASTHelper.packExpression(baseExp), methodName, exp, args);
		
		logger.endOfMethod(result);
		return result;
	}
	
	private OclExpressionCS generateCollWithTuple(OclExpressionCS baseExp, 
			Example target, LinkExample link, boolean alreadyVisited) throws Exception {		
		logger.startOfMethod("generateCollWithTuple");
			
		// assertions
		if ( !((link == null) || (link.getTarget() == target)) ) 
			throw new Exception("Invalid parameters of " + getClass().getName() + ".generateCollWithTuple() method!");
		if ( (link == null) && alreadyVisited ) 
			throw new Exception("Invalid parameters of " + getClass().getName() + ".generateCollWithTuple() method!");

		boolean linkToNestedOutput = (link != null) && model.linkToNestedOutput(link); 
		boolean linkToForAll = (link != null) && model.linkToForAll(link);
		
		logger.printLogLine("linkToNestedOutput: " + linkToNestedOutput);
		logger.printLogLine("linkToForAll: " + linkToForAll);
		
		// ->collect(vTemp | Tuple {v1 = v, v2 = vTemp})
		// ->collect(vTemp | Tuple {v1 = v.v1, ..., v2 = vTemp})
		
		// iterator
		VariableDeclarationCS iteratorDecl = new VariableDeclarationCS(ASTHelper.createSN("vTemp"));
		ArgumentsCS args = new ArgumentsCS(iteratorDecl);
		// collection method name
		SimpleNameCS methodName = ASTHelper.createSN("collect");
		// expression - tuple
		SimpleNameCS iteratorName = ASTHelper.createSN("v");
		VariableDeclarationListCS vars = new VariableDeclarationListCS();			
					
		// first collect
		if (prevTuple == null) {
			if (!linkToForAll) { 
				Element tmpUmlElement;
				// first field: v1 = v
				SimpleNameCS varName = ASTHelper.createSN("v1");
				tmpUmlElement = link != null ? link.getSource().getDescribes() : target.getDescribes();
				AttributeOrVariableCS initValue1 = ASTHelper.createAttrOrVar(iteratorName, tmpUmlElement);
				
				VariableDeclarationCS varDecl = new VariableDeclarationCS(varName, initValue1);
				varDecl.umlElement = link != null ? link.getDescribedBy() : target.getDescribes();
				vars.declarations.add(varDecl);
				
				outputs.replaceASTNode(startingExp, varDecl);
				sorts.replaceASTNode(startingExp, varDecl);
				if (link != null)
					fields.put(link.getSource(), varName.value);
				else
					fields.put(firstExample, varName.value); // missing first example->v1
			}
		// next collects
		} else {
			// rewrite previous fields
			for (VariableDeclarationCS var : prevTuple.varDeclList.declarations) {
				SimpleNameCS varName = ASTHelper.createSN(var.name.value);
				AttributeCallCS initValue1 = ASTHelper.createAttrCall(iteratorName, var.name.value, var.umlElement);				
				
				VariableDeclarationCS varDecl = new VariableDeclarationCS(varName, initValue1);
				varDecl.umlElement = var.umlElement;
				vars.declarations.add(varDecl);
				
				if (!alreadyVisited) {
					outputs.replaceASTNode(var, varDecl);
					sorts.replaceASTNode(var, varDecl);
				}
			}
		}
		
		// add new field		
		SimpleNameCS varName = ASTHelper.createSN("v" + (vars.declarations.size() + 1));
		OclExpressionCS /*AttributeOrVariableCS or AttributeCallCS*/ lastAttr;
		VariableDeclarationCS varDecl;
		if (linkToNestedOutput) {
			lastAttr = generateLinkExpression(link);
			varDecl = new VariableDeclarationCS(varName, (AttributeCallCS) lastAttr);
		} else {
			lastAttr = ASTHelper.createAttrOrVar("vTemp", target.getDescribes());
			varDecl = new VariableDeclarationCS(varName, (AttributeOrVariableCS) lastAttr);
		}			
		vars.declarations.add(varDecl);
		vars.updateChildPointers();				
		
		if (!alreadyVisited) {
			processPredicates(target, varDecl);
			if (target.getOutput() != null)
				addOutput(target, varDecl, null, target.getOutput().getTupleFieldName(), false, null);
			fields.put(target, varName.value);
		}
	
		// collect expression
		TupleLiteralExpCS tuple = new TupleLiteralExpCS(vars);	
		prevTuple = tuple;
		OclExpressionCS result = new CollectionOperationCS(
				ASTHelper.packExpression(baseExp), methodName, tuple, args);
			
		logger.endOfMethod(result);
		return result;				
	}
	
	private OclExpressionCS generateForAll(OclExpressionCS baseExp, LinkExample link) throws Exception {	
		logger.startOfMethod("generateForAll");
				
		ArrayList<VariableDeclarationCS> vars = prevTuple.varDeclList.declarations;
		int varCount = vars.size();
		String prevVarName = vars.get(varCount - 1).name.value; 
		
		// nested expression
		// for simple for-all: vn.linkName
		ASTExporter exporter = new ASTExporter();				
		OclExpressionCS nestedExp = (OclExpressionCS) exporter.generate(
				model, new GenOptions(link, prevVarName, CallType.FOR_ALL)
				);
		logger.message("nested expression", nestedExp);
		
		// ->select( <nestedExp>->collect(vTemp | Tuple {vn+1 = vTemp})-> forAll(filterExp) )->collect(vn)
		
		// ->collect
		OclExpressionCS exp = ASTHelper.createAttrOrVar("vTemp", null);
		SimpleNameCS fieldName = ASTHelper.createSN("v" + (varCount + 1)); 
		OclExpressionCS initValue = ASTHelper.createAttrOrVar("vTemp", null);
		VariableDeclarationCS varDecl = new VariableDeclarationCS(fieldName, initValue);
		VariableDeclarationListCS varDecls = new VariableDeclarationListCS(varDecl, null); 
		TupleLiteralExpCS tuple = new TupleLiteralExpCS(varDecls);
		ArgumentsCS args = new ArgumentsCS(exp);
		OclExpressionCS result = new CollectionOperationCS(
				nestedExp, ASTHelper.createSN("collect"), tuple, args
				);		
			
		// -> forAll
		result = generateFilterExp(result, link.getTarget(), "v" + (varCount + 1));
		
		// ->select
		args = new ArgumentsCS(result);
		result = new CollectionOperationCS(baseExp, new SimpleNameCS("select"), args);		
	
		// ->collect(vn)				
		AttributeOrVariableCS vn = ASTHelper.createAttrOrVar(prevVarName, null);
		result = new CollectionOperationCS(result, new SimpleNameCS("collect"), vn, null);		
				
		logger.endOfMethod(result);	
		return result;
	}
	
	private OclExpressionCS generateVisited(OclExpressionCS baseExp, LinkExample link) throws Exception {
		logger.startOfMethod("generateVisited");
		
		String rVar = fields.get(link.getTarget());
		String lVar = getLastVarName((CollectionOperationCS) baseExp);
		AttributeOrVariableCS exp1 = ASTHelper.createAttrOrVar(lVar, null /* not needed */);
		AttributeOrVariableCS exp2 = ASTHelper.createAttrOrVar(rVar, null);		
		BinaryExpressionCS expression = new BinaryExpressionCS(exp1, exp2, Operator.opEquals);
		CollectionOperationCS result = new CollectionOperationCS(baseExp, 
				ASTHelper.createSN("select"), expression, null);
		
		logger.endOfMethod(result);
		return result;
	}
	
	protected OclExpressionCS generateSortExp(OclExpressionCS baseExp) throws Exception {
		logger.startOfMethod("generateSortExp");

		OclExpressionCS result = baseExp;
		if (sorts.size() > 0) {
			sorts.sort();			
			for (OQBESort item : sorts) {
				SimpleNameCS methodName = ASTHelper.createSN("sortedBy");
				OclExpressionCS exp = generateOutValue(item, null);
				if (item.sortFlag.getOrder() == SortOrder.DESCENDING_LITERAL)
					exp = new UnaryExpressionCS(Operator.opMinus, exp);
				result = new CollectionOperationCS(result, methodName, exp, null);
			}
		}		
		
		logger.endOfMethod(result);
		return result;
	}
	
	protected void generateNestedOutputs() throws Exception {
		for (OQBEOutput output : outputs) {
			if (output.nestedOutput) {
				ASTExporter exporter = new ASTExporter();				
				AbstractNode astNode = exporter.generate(model,
						new GenOptions(output.nestedLink, output.fieldName, CallType.NESTED_OUTPUT));
				output.astNode = astNode;
			}
		}
	}
	
	protected OclExpressionCS generateFinalCollect(OclExpressionCS baseExp) throws Exception {
		logger.startOfMethod("generateFinalCollect");		
		
		OclExpressionCS result = baseExp;
		SimpleNameCS opName = ASTHelper.createSN("collect");		
		if (outputs.size() == 0) {
			if (options.callType == CallType.STANDARD)
				throw new Exception("No outputs defined!");
		} else if (outputs.size() == 1) {
			OQBEItem output = outputs.get(0);
			// do not generate collect operation for outputs in starting example without tuple field name
			if ( !((output.astNode == startingExp) && (output.fieldName == null)) ) {
				OclExpressionCS exp = generateOutValue(outputs.get(0), null);
				result = new CollectionOperationCS(ASTHelper.packExpression(result), opName, exp, null);
			}
		} else {
			VariableDeclarationListCS vars = new VariableDeclarationListCS();
			for (OQBEOutput output : outputs) {
				validateOutputName(output);
				SimpleNameCS varName = ASTHelper.createSN(output.outName);
				OclExpressionCS initValue = generateOutValue(output, "t");
				VariableDeclarationCS var = new VariableDeclarationCS(varName, initValue);
				vars.declarations.add(var);
			}
			vars.updateChildPointers();
			TupleLiteralExpCS tuple = new TupleLiteralExpCS(vars);
			VariableDeclarationCS iteratorDecl = new VariableDeclarationCS(ASTHelper.createSN("t"));
			ArgumentsCS args = new ArgumentsCS(iteratorDecl);
			result = new CollectionOperationCS(ASTHelper.packExpression(result), opName, tuple, args);
		}
		
		logger.endOfMethod(result);
		return result;
	}

	private OclExpressionCS generateComparators(OclExpressionCS baseExp) throws Exception {
		logger.startOfMethod("generateComparators");
		
		OclExpressionCS result = baseExp;
		for (Comparator comp : model.getComparators()) {
			Operator op = Operator.getOperator(comp.getOperator());
			
			OclExpressionCS lExp = generateLRExp(comp.getLeftHandExp());
			OclExpressionCS rExp = generateLRExp(comp.getRightHandExp());						
			BinaryExpressionCS binExp = new BinaryExpressionCS(lExp, rExp, op);
			
			SimpleNameCS collOpName = ASTHelper.createSN("select");
			result = new CollectionOperationCS(result, collOpName, binExp, null);
		}
		
		logger.endOfMethod(result);
		return result;
	}	
	
	private OclExpressionCS generateLRExp(OQBE.Comparable comp) {
		assert( (comp instanceof Example) || (comp instanceof Predicate) );

		OclExpressionCS result = null;
		
		Example example;
		if (comp instanceof Example)
			example = (Example) comp;
		else 
			example = model.getParentExample((Predicate) comp);
		String exampleName = fields.get(example);
		if (exampleName != null)
			result = ASTHelper.createAttrOrVar(ASTHelper.createSN(exampleName), example.getDescribes());
		
		if (comp instanceof Predicate) {
			Predicate pred = (Predicate) comp;
			AttributeOrVariableCS attr = ASTHelper.createAttrOrVar(pred.getPredicateName(), pred.getSubject());
			if (result != null)
				result = new AttributeCallCS(result, attr);
			else
				result = attr;
		}
		
		return result;
	}	
		
	private OclExpressionCS generateOutValue(OQBEItem output, String iteratorName) throws Exception {
				
		// nested output
		if ((output instanceof OQBEOutput) && ((OQBEOutput) output).nestedOutput)
			return (OclExpressionCS) output.astNode;
		else {
			// standard tuple field
			if ( (output.astNode instanceof VariableDeclarationCS) ||
			     (output.astNode instanceof AttributeOrVariableCS) //||
				 ) { //(output.astNode instanceof AttributeCallCS) ) {
				NotNullEList<String> elements = new NotNullEList<String>();				
				elements.add(iteratorName);		
				if (output.astNode != startingExp) {
					if (output.astNode instanceof VariableDeclarationCS)
						elements.add(((VariableDeclarationCS) output.astNode).name.value);
					else if (output.astNode instanceof AttributeOrVariableCS)
						elements.add(((AttributeOrVariableCS) output.astNode).name.value);
				}
				elements.add(output.fieldName);				
				// create attribute/variable or attribute call
				return ASTHelper.createVoAoAC(elements);
			} 
			// expression
			else {					
				Example parentExample = model.getParentExample((OclExpression) output.diagramItem);
				String fieldName = fields.get(parentExample);
				if (fieldName != null) {
					// additional Bag{<fieldName>}-> for output expressions from previous examples 
					// e.g. collect(Bag {v2}->collect(quantity * product.price));		
					NotNullEList<String> elements = new NotNullEList<String>();
					elements.add(iteratorName);
					elements.add(fieldName);
					OclExpressionCS exp = ASTHelper.createVoAoAC(elements); 
					CollectionLiteralExpCS collLiteral = ASTHelper.createBagLiteral(exp);
					//ArgumentsCS arguments = new ArgumentsCS((OclExpressionCS) output.astNode);
					//return new CollectionOperationCS(collLiteral, ASTHelper.createSN("collect"), arguments);
					return new CollectionOperationCS(collLiteral, ASTHelper.createSN("collect"), (OclExpressionCS) output.astNode, null);
				} else
					return (OclExpressionCS) output.astNode;
			}
		}
	}

	// process
	
	private void processPredicates(Example example, AbstractNode astNode) throws Exception {
		for (Object predicate : example.getPredicates())
			processPredicate((Predicate) predicate, astNode);
	}

	private void processPredicate(Predicate predicate, AbstractNode astNode) throws Exception {
		if (predicate.getOutput() != null)
			addOutput(predicate, astNode, getPredicateName(predicate), predicate.getOutput().getTupleFieldName(), false, null);
		if (predicate.getSort() != null)
			addSort(predicate, astNode, getPredicateName(predicate), predicate.getSort());		
	}
	
	private void processOclExpressions(Example example) throws Exception {
		for (Object oclExpression : example.getExpressions())
			processOclExpression((OclExpression) oclExpression);
	}
	
	private void processOclExpression(OclExpression oclExpression) throws Exception {
		String fieldName = getOclExpressionOutput(oclExpression);
		OclExpressionCS generatedExp = parseExp(oclExpression);;
		if (oclExpression.getOutput() != null)
			addOutput(oclExpression, generatedExp, fieldName, oclExpression.getOutput().getTupleFieldName(), false, null);
		if (oclExpression.getSort() != null)
			addSort(oclExpression, generatedExp, fieldName, oclExpression.getSort());
	}

	// add
			
	private void addSort(EObject diagramItem, AbstractNode astNode, String fieldName, SortFlag sortFlag) {
		sorts.add(new OQBESort(diagramItem, astNode, fieldName, sortFlag));
	}	

	private void addOutput(EObject diagramItem, AbstractNode astNode, String fieldName, String outName, Boolean nestedOutput, LinkExample nestedLink) {
		outputs.add(new OQBEOutput(diagramItem, astNode, fieldName, outName, nestedOutput, nestedLink));
	}

	// get
	
	private Example getFirstExample() throws Exception {		
		if (model.getExamples().size() == 0)
			throw new Exception("Empty OQBE diagram!");
		Example result = null;
		EList<Example> firstExamples = model.examplesWithoutIncomingLinks();
		if (firstExamples.size() == 0)
			result= model.getExamples().get(0);			
		else  
			result = firstExamples.get(0);
		return result;
	}
	
	private String getExampleName(Example example) throws Exception {
		if ((example.getExampleName() != null) && (example.getExampleName() != ""))
			return example.getExampleName();
		else {
			if (example.getDescribes() == null) {
				if (debugMode) {
					if ( (example.getClassName() == null) || (example.getClassName() == "") )
						throw new Exception("Empty class name!");
					else
						return example.getClassName();
				} else
					throw new Exception("Empty \"describes\" property!");
			} else
				return example.getDescribes().getName();
		}
	}

	private String getPredicateName(Predicate predicate) throws Exception {
		if (predicate.getSubject() != null)
			return predicate.getSubject().getName();
		else {
			if ( debugMode && (predicate.getPredicateName() != null) && (predicate.getPredicateName() != "") )
				return predicate.getPredicateName();
			else
				throw new Exception("Empty subject!\n" + "Predicate: "
						+ predicate.getPredicateName() + ".");
		}
	}
	
	private String getLinkName(LinkExample link) throws Exception {
		if (link.getDescribedBy() != null)
			return link.getDescribedBy().getName();
		else {
			if ( debugMode && (link.getLinkName() != null) && (link.getLinkName() != "") )
				return link.getLinkName();
			else
				throw new Exception("Empty link name!\n" + "Example: "
						+ link.getSource().getExampleName() + ".");
		}
	}

	private String getLastVarName(CollectionOperationCS collection) {
		if (collection.expression instanceof CollectionOperationCS)
			collection = (CollectionOperationCS) collection.expression;
		TupleLiteralExpCS parentTuple = (TupleLiteralExpCS) collection.expression;
		ArrayList<VariableDeclarationCS> vars = parentTuple.varDeclList.declarations;
		return vars.get(vars.size() - 1).name.value;
	}

	private String getOclExpressionOutput(OclExpression oclExpression) {
		try {
			BinaryExpressionCS expression = (BinaryExpressionCS) ((VideEditorCode) VideCodeToAST.convert(oclExpression.getExpression() + ";")).statements.get(0);
			if (expression.expression1 instanceof AttributeOrVariableCS) {
				return ((AttributeOrVariableCS) expression.expression1).name.value;
			}
		} catch (Exception e) {
		}		
		return null;
	}
	
	private EList<Predicate> getNotFlaggedPredicates(Example example) {
		EList<Predicate> result = new BasicEList<Predicate>();
		for (Object obj : example.getPredicates()) {
			Predicate predicate = (Predicate) obj;
			// check output and sort flags
			if ( (predicate.getOutput() == null) && (predicate.getSort() == null) ) {
				// check comparators
				boolean ok = true;
				EList<Comparator> comps = model.getComparators();
				int idx = 0;
				while ( ok && (idx < comps.size()) ) {
					Comparator comp = comps.get(idx);
					if ( (comp.getLeftHandExp() == predicate) || (comp.getRightHandExp() == predicate) )
						ok = false;
					else
						idx ++;
				}
				if (ok)
					result.add(predicate);
			}
		}
		return result;
	}
	
	private EList<OclExpression> getNotFlaggedOclExpressions(Example example) {
		EList<OclExpression> result = new BasicEList<OclExpression>();
		for (Object obj : example.getExpressions()) {
			OclExpression oclExpression = (OclExpression) obj;
			if ( (oclExpression.getExpression() != null) && 
				 (oclExpression.getExpression() != "") &&
				 (oclExpression.getOutput() == null) &&
				 (oclExpression.getSort() == null) )
				result.add(oclExpression);
		}		
		return result;
	}

	// other
	
	protected boolean singleNestedOutput(OQBEItem output) {
		// get (parent) example
		Example example;
		if (output.diagramItem instanceof Example)
			example = (Example) output.diagramItem;
		else if (output.diagramItem instanceof Predicate) {
			Predicate predicate = (Predicate) output.diagramItem;
			example = model.getParentExample(predicate);
		} else {
			OclExpression oclExp = (OclExpression) output.diagramItem;
			example = model.getParentExample(oclExp);
		}
		boolean result = false;
		// get parent region (nested output)
		NestedOutput region = model.getNestedOutput(example);
		if (region != null) {
			result = true;
			// check output count
			@SuppressWarnings("unchecked")
			EList<Example> nestedExamples = (EList<Example>) region.getNestedExamples();
			for (int exIdx = 0; (exIdx < nestedExamples.size()) && result; exIdx ++) {
				example = nestedExamples.get(0);
				if ( (example.getOutput() != null) && (example != output.diagramItem) )
					result = false;
				@SuppressWarnings("unchecked")
				EList<Predicate> predicates = (EList<Predicate>) example.getPredicates();
				for (int prIdx = 0; (prIdx < predicates.size()) && result; prIdx ++) {
					Predicate predicate = predicates.get(prIdx);
					if ( (predicate.getOutput() != null) && (predicate != output.diagramItem) )
						result = false;
				}
			}
		}
		return result;
	}

	private void validateOutputName(OQBEOutput output) throws Exception {
		if ( (output.outName == null) || (output.outName == "") ) {
			if (!singleNestedOutput(output)) {
				String diagramItemName = "";
				if (output.diagramItem instanceof Example)
					diagramItemName += "Example: " + getExampleName((Example) output.diagramItem);
				else if (output.diagramItem instanceof Predicate)
					diagramItemName += "Predicate: " + getPredicateName((Predicate) output.diagramItem);
				else if (output.diagramItem instanceof OclExpression) {
					diagramItemName += "OCL Expression: " + output.astNode.asCode();
					if (diagramItemName.endsWith(";"))
						diagramItemName = diagramItemName.substring(0, diagramItemName.length() - 1);
				}
				if (diagramItemName != "")
					diagramItemName = "\n" + diagramItemName + ".";
				throw new Exception("Missing output name!" + diagramItemName);
			}
		}
	}
	
	protected void rememberNestedOutput(LinkExample link) throws Exception {
		Example target = link.getTarget();
		NestedOutput output = model.getNestedOutput(target);
		String outName = output.getTupleFieldName();
		String fieldName = ASTHelper.lastVarName(prevTuple);
		OQBEOutput outputItem = new OQBEOutput(target, null, fieldName, outName, true, link);
		validateOutputName(outputItem);
		outputs.add(outputItem);
	}
	
}
