package oqbe.exporters;

import org.eclipse.core.runtime.IPath;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.ui.part.FileEditorInput;

import OQBE.Comparator;
import OQBE.DiagramObject;
import OQBE.DiagramPackage;
import OQBE.Example;
import OQBE.ForAll;
import OQBE.LinkExample;
import OQBE.NestedOutput;
import OQBE.OclExpression;
import OQBE.Predicate;


public class OQBEModel {

	// required plug-ins (plugin.xml -> Dependencies):
	// 	- org.eclipse.emf.workspace
	//  - org.eclipse.gmf.runtime.notation
	
	private DiagramPackage pckg = null;
	private EList<Example> examples = null;
	private EList<DiagramObject> objects = null;
	private EList<LinkExample> links = null;

	public OQBEModel(IPath path) throws Exception {
		if (path.getFileExtension().equals("oqbe"))
			prepare(getDiagramPackage(path.toOSString()));
		else if (path.getFileExtension().equals("oqbe_diagram")) {
			Diagram diagram = getDiagram(path.toOSString());
			prepare(getDiagramPackage(diagram));
		} else 
			throw new Exception("Invalid editor input!\n" + 
					"Select OQBE diagram or model file to generate\n" + 
					"VIDE source code file.");
	}

	public OQBEModel(FileEditorInput editorInput) throws Exception {
		this(editorInput.getPath());
	}
	
	//
	
	@SuppressWarnings("unchecked")	
	
	private void prepare(DiagramPackage diagramPackage) {
		pckg = diagramPackage;
		examples = pckg.getExamples();
		objects = pckg.getObjects();
		links = new BasicEList<LinkExample>();
		for (DiagramObject obj : objects)
			if (obj instanceof LinkExample)
				links.add((LinkExample) obj);		
	}
	
	//
		
	public static DiagramPackage getDiagramPackage(String modelFileName) {
		ResourceSet resSet = new ResourceSetImpl();
		URI fileURI = URI.createFileURI(modelFileName);	
		Resource res = resSet.getResource(fileURI, true);
		return (DiagramPackage) res.getContents().get(0);		 
	}

	public static DiagramPackage getDiagramPackage(Diagram diagram) {		
		return (DiagramPackage) diagram.getElement();
	}
		
	public static Diagram getDiagram(String diagramFileName) {
		ResourceSet resSet = new ResourceSetImpl();
		URI fileURI = URI.createFileURI(diagramFileName);	
		Resource res = resSet.getResource(fileURI, true);	
		return (Diagram) res.getContents().get(0);			
	}
				
	//

	public EList<LinkExample> getLinkedExamples(Example source) {
		EList<LinkExample> result = new BasicEList<LinkExample>();
		for (LinkExample link : links)
			if (link.getSource() == source)
				result.add(link);				
		return result;		
	}
	
	public EList<LinkExample> getLinks(Example source, Example target) {
		EList<LinkExample> result = new BasicEList<LinkExample>();
		for (LinkExample link : links)
			if ( (link.getSource() == source) && (link.getTarget() == target) )
				result.add(link);				
		return result;		
	}
	
	public EList<Example> examplesWithoutIncomingLinks() {
		EList<Example> result = new BasicEList<Example>();
		for (Example example : examples)
			result.add(example);
		for (LinkExample link : links)
			result.remove(link.getTarget());
		return result;		
	}
	
	public EList<Comparator> getComparators() {	
		EList<Comparator> result = new BasicEList<Comparator>(); 
		for (Object obj : pckg.getObjects())
			if (obj instanceof Comparator)
				result.add((Comparator) obj);
		return result;
	}
	
	//
	
	public DiagramPackage getPackage() {
		return pckg;
	}
	
	public EList<Example> getExamples() {	
		return examples;		
	}
	
	public EList<LinkExample> getLinks() {
		return links;		
	}

	// for alls
	
	/*
	public ForAll getParentForAll(NestedOutput output, Example example) {		
		ForAll result = null;
		@SuppressWarnings("unchecked")
		EList forAlls = pckg.getForAlls();
		int idx = 0;
		while ( (result == null) && (idx < forAlls.size()) ) {
			ForAll forAll = (ForAll) forAlls.get(idx);
			if (forAll.getForAllExamples().contains(example))
				result = forAll;
			idx ++;
		}
		return result;
	}
	*/
		
	public ForAll getParentForAll(NestedOutput parentOutput, Example example) {
		ForAll result = null;
		@SuppressWarnings("unchecked")
		EList forAlls = parentOutput.getNestedForAlls();
		int idx = 0;
		while ( (result == null) && (idx < forAlls.size()) ) {
			ForAll forAll = (ForAll) forAlls.get(idx);
			if (forAll.getForAllExamples().contains(example))
				result = forAll;
			idx ++;
		}
		if (result == null) {
			@SuppressWarnings("unchecked")
			EList outputs = parentOutput.getNestedOutputs();
			idx = 0;		
			while ( (result == null) && (idx < outputs.size()) ) {
				result = getParentForAll((NestedOutput) outputs.get(idx), example);
				idx ++;;
			}
		}		
		return result;
	}

	public ForAll getParentForAll(Example example) {		
		ForAll result = null;
		// for-alls in package
		@SuppressWarnings("unchecked")	
		EList forAlls = pckg.getForAlls();
		int idx = 0;
		while ( (result == null) && (idx < forAlls.size()) ) {
			ForAll forAll = (ForAll) forAlls.get(idx);
			if (forAll.getForAllExamples().contains(example))
				result = forAll;
			else
				idx ++;
		}
		// for-alls in nested outputs of package
		if (result == null) {
			@SuppressWarnings("unchecked")
			EList outputs = pckg.getOutputs();
			idx = 0;
			while ( (result == null) && (idx < outputs.size()) ) {
				NestedOutput output = (NestedOutput) outputs.get(idx);
				result = getParentForAll(output, example);
				idx ++;
			}			
		}
		return result;
	}

	public boolean linkToForAll(LinkExample link) {
		if (link == null)
			return false;
		else {
			ForAll parentBlockOfSrc = getParentForAll(link.getSource());		
			ForAll parentBlockOfDst = getParentForAll(link.getTarget());
			return parentBlockOfDst != parentBlockOfSrc;
		}
	}

	// nested outputs
	
	public NestedOutput getNestedOutput(Example example) {		
		NestedOutput result = null;
		@SuppressWarnings("unchecked")
		EList outputs = pckg.getOutputs();
		int idx = 0;
		while ( (result == null) && (idx < outputs.size()) ) {
			NestedOutput output = (NestedOutput) outputs.get(idx);
			if (output.getNestedExamples().contains(example))
				result = output;
			else
				result = getNestedOutput(output, example); 
			idx ++;
		}
		return result;
	}
	
	protected NestedOutput getNestedOutput(NestedOutput parentOutput, Example example) {
		NestedOutput result = null;
		@SuppressWarnings("unchecked")
		EList outputs = parentOutput.getNestedOutputs();
		int idx = 0;
		while ( (result == null) && (idx < outputs.size()) ) {
			NestedOutput output = (NestedOutput) outputs.get(idx);
			if (output.getNestedExamples().contains(example))
				result = output;
			else
				result = getNestedOutput(output, example);
			idx ++;
		}
		return result;	
	}
	
	public boolean linkToNestedOutput(LinkExample link) {
		if (link == null)
			return false;
		else {
			NestedOutput parentBlockOfSrc = getNestedOutput(link.getSource());		
			NestedOutput parentBlockOfDst = getNestedOutput(link.getTarget());
			return parentBlockOfDst != parentBlockOfSrc;
		}
	}
	
	// parent example
	
	public Example getParentExample(Predicate predicate) {
		Example result = getParentExample(predicate, pckg.getExamples());
		if (result == null) {
			@SuppressWarnings("unchecked")		
			EList<NestedOutput> outputs = (EList<NestedOutput>) pckg.getOutputs();
			for (int idx = 0; (idx < outputs.size()) && (result == null); idx ++)
				result = getParentExample(predicate, outputs.get(idx).getNestedExamples());
		}
		return result;
	}
	
	@SuppressWarnings("unchecked")
	
	protected Example getParentExample(Predicate predicate, EList examples) {
		Example result = null;
		for (int idx = 0; (idx < examples.size()) && (result == null); idx ++) {
			Example example = (Example) examples.get(idx);
			if (example.getPredicates().contains(predicate))
				result = example;
		}
		return result;		
	}

	@SuppressWarnings("unchecked")
	
	protected Example getParentExample(OclExpression oclExp, EList examples) {
		Example result = null;
		for (int idx = 0; (idx < examples.size()) && (result == null); idx ++) {
			Example example = (Example) examples.get(idx);
			if (example.getExpressions().contains(oclExp))
				result = example;
		}
		return result;		
	}
	
	public Example getParentExample(OclExpression oclExp) {
		Example result = getParentExample(oclExp, pckg.getExamples());
		if (result == null) {
			@SuppressWarnings("unchecked")		
			EList<NestedOutput> outputs = (EList<NestedOutput>) pckg.getOutputs();
			for (int idx = 0; (idx < outputs.size()) && (result == null); idx ++)
				result = getParentExample(oclExp, outputs.get(idx).getNestedExamples());
		}
		return result;
	}
	
}
